/******************************************************************************
    Copyright (c) 2007 Electronic Tools Company    ALL RIGHTS RESERVED

  The contents of this file are subject to the restrictions and limitations
  set forth in the SYNOPSYS Open Source License Version 1.0  (the "License"); 
  you may not use this file except in compliance with such restrictions 
  and limitations. You may obtain instructions on how to receive a copy of 
  the License at

  http://www.synopsys.com/partners/tapin/open_source/license_liberty.html. 

  Software distributed by Original Contributor under the License is 
  distributed on an "AS IS" basis, WITHOUT WARRANTY OF ANY KIND, either 
  expressed or implied. See the License for the specific language governing 
  rights and limitations under the License.

******************************************************************************/

#include "stdc++.h"
#include "ccs2ecsm.h"
#include "library.h"


int main(int argc, char ** argv)
{
  static string strInput, strOutput, strCap, strVol;

  if ( argc < 2) {
    ShowHelp();
    return 1;
  }

  if ((0==strcasecmp(argv[1], "-h")) || (0==strcasecmp(argv[1], "-help")))
    {
      ShowHelp();
      return 1;
    }

  if (argc%2==0)
    {
      cerr<<"Error: Please provide \"-\" option name and value in pair."<<endl<<endl;
      ShowHelp();
      return 2;
    }

  map <string , string> opt;

  for (int i=1; i<argc-1; i+=2)
    opt[argv[i]]=argv[i+1];

  if (opt.find("-input")!=opt.end())
    strInput=opt["-input"];
  if (opt.find("-output")!=opt.end())
    strOutput=opt["-output"];
  if (opt.find("-cap_convert_method")!=opt.end())
    strCap=opt["-cap_convert_method"];
  if (opt.find("-voltage_sample_point")!=opt.end())
    strVol=opt["-voltage_sampel_point"];

  if (0==strInput.size())
    {
      cerr<<"Error: Please provide input ccs libname with -input ."<<endl<<endl;
      ShowHelp();
      return 3;
    }
  
  if (0==strOutput.size())
    {
      strOutput=strInput;
      int i=strOutput.find_last_of(".");
      strOutput.insert(i, "ecsm");
      cout<<"No output filename provided by command line option, use "<<strOutput<<" as default."<<endl;
    }
  else if (strInput==strOutput)
    {
      strOutput=strInput;
      int i=strOutput.find_last_of(".");
      strOutput.insert(i, "ecsm");
      cerr<<"Warning: Same file specified both as Input and Output. Default output filename "<<strOutput<<" is used instead, to avoid overwrite of input .lib."<<endl;
    }

  if (0==strCap.size())
    strCap="first";
  if ((strCap!= "ave") && (strCap!= "max") && (strCap!= "min") && (strCap!= "first") && (strCap!= "second"))
    {
      cerr<<"Warning: Legal options are \"ave\",\"max\",\"min\",\"first\",\"second\", not "<<strCap<<"."<<endl;
      cerr<<"Use default(first) instead."<<endl;
      strCap="first";
    }

  if (0==strVol.size())
    {
      strVol="0.05,0.10,0.20,0.30,0.40,0.50,0.60,0.70,0.80,0.90,0.95";
    }

  Library lib;

  try
    {
      lib.ReadLibrary(strInput);
      lib.Ccs2ecsm(strVol, strCap);
      lib.WriteLibrary(strOutput, false);
    }
  catch(exception& e)
    {
      cerr<<e.what()<<endl;
    };

  return 0;
}

void ShowHelp()
{
  cerr<<"ccs2ecsm -help"<<endl

      <<"*******************************"<<endl
      <<"Convert CCS to ECSM "<<endl
      <<"ccs2ecsm is a utility that reads in a library that contains CCS Timing models and converts those models to the ECSM representation."<<endl<<endl
      <<"Usage: ccs2ecsm -input <inlib_name> [-output <outlib_name>] [-cap_convert_method <...>] [-voltage_sample_point <...>]"<<endl<<endl
      <<"-input inlib_name    Specifies the name of the input library containing the CCS models."<<endl<<endl
      <<"-output outlib_name  Specifies the name of the file to which the modified library, containing ECSM models, should be written. The default is basename(inlib_name)+ecsm.ext(inlib_name)"<<endl<<endl
      <<"-cap_convert_method <ave | max | min | first | second>   Specifies the method to use to convert the CCS receiver model to the ECSM receiver model."<<endl<<endl
      <<"-voltage_sample_point <list> Specifies the normalized voltage sample points that should be output for the ECSM waveform. The default values are \"0.05,0.10,0.20,0.30,0.40,0.50,0.60,0.70,0.80,0.90,0.95\"."<<endl<<endl
  <<"-help		Prints this help message."<<endl;
  return;
}
