/******************************************************************************
    Copyright (c) 2007 Electronic Tools Company    ALL RIGHTS RESERVED

  The contents of this file are subject to the restrictions and limitations
  set forth in the SYNOPSYS Open Source License Version 1.0  (the "License"); 
  you may not use this file except in compliance with such restrictions 
  and limitations. You may obtain instructions on how to receive a copy of 
  the License at

  http://www.synopsys.com/partners/tapin/open_source/license_liberty.html. 

  Software distributed by Original Contributor under the License is 
  distributed on an "AS IS" basis, WITHOUT WARRANTY OF ANY KIND, either 
  expressed or implied. See the License for the specific language governing 
  rights and limitations under the License.

******************************************************************************/

#include "stdc++.h"

#include "attr.h"
#include "group.h"
#include "template.h"
#include "cell.h"
#include "pin.h"
#include "path.h"
#include "value.h"
#include "myregex.h"
#include "util.h"
#include "library.h"
#include "comment.h"
#include "empty.h"

map <string, Classtype> Group::classtype;

Group::Group(int lineno, const string& type, const string& name, Group *p)
    : Entry(lineno), strType(type), strName(name), Parent(p)
{
  init();
}

Group::~Group()
{
  for (int i=0; i<entries.size(); i++) {
    if (entries[i] != NULL) {
      delete entries[i];
      entries[i] = NULL;
    }
  }
}

void Group::init()
{
  if (classtype.empty()) {
     classtype["library"] = eLibrary;
     classtype["lu_table_template"] = eTemplate;
     classtype["power_lut_template"] = eTemplate;
     classtype["output_current_template"] = eTemplate;
     classtype["iv_lut_template"] = eTemplate;
     classtype["scaling_factors"] = eScaling_factor;
     classtype["cell"] = eCell;
     classtype["test_cell"] = eCell;
     classtype["pin"] = ePin;
     classtype["timing"] = ePath;
     classtype["internal_power"] = ePath;
     classtype["receiver_capacitance"] = ePath;
     classtype["cell_rise"] = eValue;
     classtype["cell_fall"] = eValue;
     classtype["rise_transition"] = eValue;
     classtype["fall_transition"] = eValue;
     classtype["rise_constraint"] = eValue;
     classtype["fall_constraint"] = eValue;
     classtype["rise_power"] = eValue;
     classtype["fall_power"] = eValue;
     classtype["power"] = eValue;
     classtype["vector"] = eValue;
     classtype["receiver_capacitance1_rise"] = eValue;
     classtype["receiver_capacitance1_fall"] = eValue;
     classtype["receiver_capacitance2_rise"] = eValue;
     classtype["receiver_capacitance2_fall"] = eValue;
  }
}

void Group::read(si2drGroupIdT group, char *indent)
{
  char indent2[200];
  si2drErrorT err;
  si2drNamesIdT names;
  si2drGroupsIdT groups;
  si2drAttrsIdT attrs;
  si2drDefinesIdT defs;
  si2drGroupIdT g1;
  si2drAttrIdT  attr;
  si2drDefineIdT def;
  int lineno = 0;
  Entry *entry; 
  char buf[1024];
  int debug = 0;
  
  sprintf(indent2, "%s\t", indent);
   // add the define class 

  defs = si2drGroupGetDefines(group, &err);
  for(def = si2drIterNextDefine(defs,&err);
  	!si2drObjectIsNull(def,&err);
  	def = si2drIterNextDefine(defs,&err))
  {
    si2drStringT nam, gnam;
    si2drValueTypeT valtype;
    char *vt;
  	
    si2drDefineGetInfo(def,&nam, &gnam, &valtype, &err);

    switch( valtype ) {
      case SI2DR_STRING:
        vt = "string";
        break;
          
      case SI2DR_FLOAT64:
        vt = "float";
        break;
          
      case SI2DR_INT32:
        vt = "integer";
        break;
      }
      
    //printf("%sdefine(%s,%s,%s);\n", indent2, nam, gnam, vt);
    //Add define class later.

    string strValue = "(";
    strValue += string(nam)+", "+string(gnam)+", "+string(vt);
    strValue += ")";
    lineno=si2drObjectGetLineNo(def, &err);
    entry = new Attr(lineno, "define", strValue, CLASS2);
    if (entry !=NULL) entries.push_back(entry);

  }
  si2drIterQuit(defs,&err);
  
  /* add attribute class */

  attrs = si2drGroupGetAttrs(group, &err);
  for(attr = si2drIterNextAttr(attrs,&err);
      !si2drObjectIsNull(attr,&err);
      attr = si2drIterNextAttr(attrs,&err))
  {
    si2drAttrTypeT at = si2drAttrGetAttrType(attr,&err);
    si2drValuesIdT vals;
    si2drValueTypeT vtype;
    si2drInt32T     intgr;
    si2drFloat64T   float64;
    si2drStringT    value, nam= si2drAttrGetName(attr,&err);
    si2drBooleanT   boolval;
    si2drExprT *expr;
    int first;
    int AttrClass;


    if( at == SI2DR_SIMPLE && 
        (!strcmp(nam,"default_operating_conditions")
         || !strcmp(nam,"default_wire_load")
         || !strcmp(nam,"default_wire_load_selection")))
      continue;
          

    sprintf(buf, "%s", nam);
    string strName(buf); 
    string strValue("");
      
    switch( at ) {
      case SI2DR_SIMPLE:
        AttrClass = CLASS1; 
        vtype = si2drSimpleAttrGetValueType(attr,&err);
        switch( vtype ) {
          case SI2DR_STRING:
              
          if (!strcmp(nam, "vil") || !strcmp(nam, "vih") || !strcmp(nam, "vimax") || !strcmp(nam, "vimin") || !strcmp(nam, "vol") || !strcmp(nam, "voh") || !strcmp(nam, "vomax") || !strcmp(nam, "vomin")) {
            value=si2drSimpleAttrGetStringValue(attr, &err);
	    sprintf(buf, "%s", value);
	    strValue = buf;
          }else if (!strcmp(nam, "key_bit") || !strcmp(nam, "address_width") || !strcmp(nam, "word_width") || !strcmp(nam, "key_seed") || !strcmp(nam, "tracks") || !strcmp(nam, "divided_by") || !strcmp(nam, "multiplied_by") || !strcmp(nam, "bit_from") || !strcmp(nam, "bit_to") || !strcmp(nam, "bit_width") ) {
	    sprintf(buf, "%d", atoi(si2drSimpleAttrGetStringValue(attr, &err)));
	    strValue = buf;
          } else {
            value=si2drSimpleAttrGetStringValue(attr, &err);
	    sprintf(buf, "%s", value);
	    strValue = buf;
          }
          break;
              
          case SI2DR_FLOAT64:
            if (!strcmp(nam, "key_bit") || !strcmp(nam, "address_width") || !strcmp(nam, "word_width") || !strcmp(nam, "key_seed") || !strcmp(nam, "tracks") || !strcmp(nam, "divided_by") || !strcmp(nam, "multiplied_by") || !strcmp(nam, "bit_from") || !strcmp(nam, "bit_to") || !strcmp(nam, "bit_width") ) {
	      sprintf(buf, "%d", (int)si2drSimpleAttrGetFloat64Value(attr,&err));
	      strValue = buf;
            } else {
	      sprintf(buf, "%1.6f", si2drSimpleAttrGetFloat64Value(attr,&err));
	      strValue = buf;
	    }
            break;
              
          case SI2DR_INT32:
	    {
	      sprintf(buf, "%d", si2drSimpleAttrGetInt32Value(attr,&err));
	      strValue = buf;
	    }
            break;
              
          case SI2DR_BOOLEAN:
            if( si2drSimpleAttrGetBooleanValue(attr,&err) ) {
	      sprintf(buf, "%s", "true");
	      strValue = buf;
            } else {
	      sprintf(buf, "%s", "false");
	      strValue = buf;
            }
            break;
          default:
	     break;
        }
        break;
          
      case SI2DR_COMPLEX:
        AttrClass = CLASS2;
        vals = si2drComplexAttrGetValues(attr,&err);
        first = 1;
	strValue = "(";
          
        for(si2drIterNextComplexValue(vals,&vtype, &intgr, &float64, &value, &boolval, &expr, &err);
          vtype != SI2DR_UNDEFINED_VALUETYPE;
          si2drIterNextComplexValue(vals,&vtype, &intgr, &float64, &value, &boolval, &expr, &err)) {
            if( !first ) {
              strValue += ",";
            }
            switch( vtype ) {
              case SI2DR_STRING:
		strValue += value;
		break;

              case SI2DR_FLOAT64:
	        sprintf(buf, "%1.6f", float64);
		strValue += buf;
                break;
                  
              case SI2DR_INT32:
	        sprintf(buf, "%d", intgr);
		strValue += buf;
                break;

              case SI2DR_BOOLEAN:
	        sprintf(buf, "%d", boolval);
		strValue += buf;
                break;
              }
            first = 0;
          }
	  strValue += ")";
          si2drIterQuit(vals,&err);
          break;
      default:
        break;
    }
    lineno=si2drObjectGetLineNo(attr, &err);
    entry = new Attr(lineno, strName, strValue, AttrClass);
    addAttr(strName, entries.size());
    if (debug) {
      cout << indent2 << strName << " " << strValue << " " << endl;
    }
    if (entry != NULL) entries.push_back(entry);
  }
  si2drIterQuit(attrs,&err);
  
  
  // add the group class

  groups = si2drGroupGetGroups(group, &err);
  for(g1 = si2drIterNextGroup(groups,&err);
      !si2drObjectIsNull(g1,&err);
      g1 = si2drIterNextGroup(groups,&err))
  {

    sprintf(buf, "%s",  si2drGroupGetGroupType(g1,&err));
    //string stype(si2drGroupGetGroupType(g1,&err));
    string stype (buf);
    si2drStringT str;
    int library_touched = 0;
  
    string sname("");
    names = si2drGroupGetNames(g1,&err);
    if(names) {
      for (str = si2drIterNextName(names,&err); str&&str[0]; str=si2drIterNextName(names, &err))
	{
	  if (str) { 
	    if (QuoteString(str)){
	      sprintf(buf, "\"%s\", ", str);
	    } else {
	      sprintf(buf, "%s, ", str);
	    }
	    sname+= buf;
	  }
	}
      if (sname.size()>2)
	sname.replace(sname.size()-2, 2, "");
	si2drIterQuit(names,&err);
    }
    
    if (debug) {
      cout << indent2 << "In the design: " << stype << "(" << sname << ")" <<endl;
    }

    lineno=si2drObjectGetLineNo(g1, &err);
    switch (getClasstype(stype)) {
      case eValue:
        entry = new Value(lineno, stype, sname, this);
        addValue(sname, entries.size());
        break;
      case eLibrary:
        name(sname);
	library_touched++;
	if (debug) {
          cout << "The library is touched" << endl;
	}
        break;

      case eTemplate:
        entry = new Template(lineno, stype, sname, this);
        addTemplate(sname, entries.size());
        break;
          
      case ePath:
        entry = new Path(lineno, stype, sname, this);
        break;

      case ePin:
        entry = new Pin(lineno, sname, this); 
        break;

      case eCell:
	entry = new Cell(lineno, stype, sname, this);
	addCell(sname, entries.size());
	break;

      default:
        entry = new Group(lineno, stype, sname, this);
    }
    if (library_touched) continue;
    if (entry != NULL) entries.push_back(entry);
    entry->read(g1, indent2);
  }
  si2drIterQuit(groups,&err);

  AfterRead();
  return;
}

int Group::Read(ifstream& is)
{
  int lineno = 0;
  char line[MAXLEN];

  while(is.getline(line, MAXLEN)) {
    lineno++;
    string str(line);

    while (str[str.size()-1] == '\\') { 
      str += "\n";

      is.getline(line, MAXLEN); 
      lineno++;
      str += line;
    }
    #ifdef _DEBUG
    //cout<<lineno+start_lineno()<<":"<<line<<endl;
    #endif

    regmatch_t matchptr[3];
    Entry *entry;

    if (Regex::Match(str.c_str(), eATTR_PAT1, matchptr)) {
      // attr : value;
      string sname(str.substr(matchptr[1].rm_so, 
                         matchptr[1].rm_eo - matchptr[1].rm_so)); 
      string svalue(str.substr(matchptr[2].rm_so, 
                         matchptr[2].rm_eo - matchptr[2].rm_so));
      entry = new Attr(lineno, sname, svalue, CLASS1);
      addAttr(sname, entries.size());
    } else if (Regex::Match(str.c_str(), eATTR_PAT2, matchptr)) {
      // attr(value);
      string sname(str.substr(matchptr[1].rm_so, 
                         matchptr[1].rm_eo - matchptr[1].rm_so)); 
      string svalue(str.substr(matchptr[2].rm_so, 
                         matchptr[2].rm_eo - matchptr[2].rm_so));
      entry = new Attr(lineno, sname, "(" + svalue + ")", CLASS2);
      addAttr(sname, entries.size());
    } else if (Regex::Match(str.c_str(), eGROUP_PAT, matchptr)) {
      // type(name) {
      string stype(str.substr(matchptr[1].rm_so, 
                         matchptr[1].rm_eo - matchptr[1].rm_so)); 
      string sname(str.substr(matchptr[2].rm_so, 
                         matchptr[2].rm_eo - matchptr[2].rm_so));

      switch (getClasstype(stype)) {
        case eValue:
          entry = new Value(start_lineno()+lineno, stype, sname, this);
          addValue(sname, entries.size());
          break;

        case eTemplate:
          entry = new Template(start_lineno()+lineno, stype, sname, this);
          addTemplate(sname, entries.size());
          break;
            
        case ePath:
          entry = new Path(start_lineno()+lineno, stype, sname, this);
          break;

        case ePin:
          entry = new Pin(start_lineno()+lineno, sname, this); 
	  break;

        case eCell:
	  entry = new Cell(start_lineno()+lineno, stype, sname, this);
	  addCell(sname, entries.size());
	  break;

        default:
          entry = new Group(start_lineno()+lineno, stype, sname, this);
      }

      lineno += entry->Read(is);
    } else if (str.find("/*") != string::npos &&
               str.find_first_not_of(" \t") == str.find("/*")) {    
      entry = new Comment(start_lineno()+lineno, str);
      lineno += entry->Read(is); 
    } else if (str.find_first_not_of(" \t") == string::npos) {
      entry = new Empty(start_lineno()+lineno);
    } else if (Regex::Match(str.c_str(), eGROUP_END_PAT, NULL)) {
      break;
    } else  {
      cout<<"Error: "<<start_lineno()+lineno<<" : "<<line<<endl;
    }

    if (entry != NULL) entries.push_back(entry);

  }

  AfterRead();
  return lineno;
}

void Group::WriteTo(ostream &os, const string& strPrefix)
{
  os<<strPrefix<<strType<<" ("<<strName<<") {"<<endl;

  writeContents(os,strPrefix + "\t");

  os<<strPrefix<<"}"<<endl;
}

void Group::writeContents(ostream& os, const string& strPrefix)
{
  for(int i=0; i<entries.size();i++) {
    if (entries[i] != NULL) {
      entries[i]->WriteTo(os, strPrefix);
    }
  }
}

Classtype Group::getClasstype(const string& stype) const
{
  map <string, Classtype>::const_iterator p = classtype.find(stype);

  if (p != classtype.end()) {
    return p->second;
  } else {
    return eGroup;
  }
}

Entry* Group::findTemplate(const string& sname) const
{
  return Parent->findTemplate(sname);
}

void Group::searchTemplates(map <string, int>& lut) const
{
  for(int i=0; i<entries.size(); i++) {
    if (entries[i] != NULL) {
      entries[i]->searchTemplates(lut);
    }  
  }
}

void Group::addAttr(const string& sname, size_t index) 
{
  attrs[sname] = index;
}

size_t Group::findAttr(const string& sname) const
{
  map <string, size_t>::const_iterator p = attrs.find(sname);
  if (p != attrs.end()) {
    return p->second;
  } else {
    throw runtime_error("invalid attribute name: " + sname);
  }
}

Entry* Group::findSubgroup(const string& stype) const
{
  for(int i=0; i<entries.size(); i++) {
    if ((entries[i]!=NULL) && (entries[i]->type() == stype)) {
      return entries[i];
    }   	
  }

  throw runtime_error("invalid group type: " + stype);
}

Entry* Group::findSubgroup(const string& stype,const string& sname) const
{
  for(int i=0; i<entries.size(); i++) {
    if (entries[i]->type() == stype && entries[i]->name() == sname) { 
      return entries[i];
    }
  }

  throw runtime_error("invalid group name: " + sname);
}

void Group::Ccs2ecsm(const string& strVol, const string& strCap)
{
  for (int i=0; i<entries.size(); ++i)
    {
      if (entries[i]!=NULL)
	entries[i]->Ccs2ecsm(strVol, strCap);
    }

  return;
}

void Group::removeSubgroup(const string& gtype)
{
  for (int i=0; i<entries.size(); ++i)
    {
      if ((entries[i]!=NULL))
	{
	  Group * g=dynamic_cast<Group *>(entries[i]);
	  if (g  && (g->type().find(gtype)!=string::npos))
	    {
	      delete entries[i];
	      entries[i]=NULL;
	    }
	}
    }
  return;
}
