/******************************************************************************
    Copyright (c) 2007 Electronic Tools Company    ALL RIGHTS RESERVED

  The contents of this file are subject to the restrictions and limitations
  set forth in the SYNOPSYS Open Source License Version 1.0  (the "License"); 
  you may not use this file except in compliance with such restrictions 
  and limitations. You may obtain instructions on how to receive a copy of 
  the License at

  http://www.synopsys.com/partners/tapin/open_source/license_liberty.html. 

  Software distributed by Original Contributor under the License is 
  distributed on an "AS IS" basis, WITHOUT WARRANTY OF ANY KIND, either 
  expressed or implied. See the License for the specific language governing 
  rights and limitations under the License.

******************************************************************************/

#include "stdc++.h"
#include "util.h"

#include "library.h"
#include "pin.h"
#include "value.h"
#include "attr.h"
#include "myregex.h"
#include "entry.h"

Library::Library(int lineno) : Group(lineno, "library", "")
{
}

Library::~Library()
{
}

void Library::addTemplate(const string& sname, size_t index)
{
  templates[sname] = index;
}

Entry* Library::findTemplate(const string& sname) const
{
  map <string, size_t>::const_iterator p = templates.find(sname);

  if (p != templates.end()) {
    return entries[p->second];
  } else {
    throw runtime_error("invalid template name: " + sname);
    return NULL;
  }
}

void Library::addCell(const string& sname, size_t index)
{
  cells[sname] = index;
}

void Library::OpenLibrary(const string& sLibname)
{
//  cout<<"Open library from '"<<sLibname<<"'"<<endl;

  try {

    ifstream inLib(sLibname.data());
    char line[MAXLEN];
    int lineno = 0;

    while (inLib.getline(line, MAXLEN)) {
      lineno++;

      regmatch_t matchptr[2];
            
      if(Regex::Match(line, eLIB_PAT, matchptr)) {
        string str(line);
        name(str.substr(matchptr[1].rm_so, 
                        matchptr[1].rm_eo - matchptr[1].rm_so));
        lineno += Read(inLib);
      } else {
        cerr<<"Syntax error: Line "<<start_lineno()+lineno<<", "
            <<line<<endl;
      }

    }
//    cout<<lineno<<" lines read in"<<endl;

  }
  catch (ios_base::failure& var) {
    cerr << var.what()<<endl;
  }

}

void Library::ReadLibrary(const string& sLibname)
{
  cout<<"Open liberty file '"<<sLibname<<"'"<<endl;

  try {
    char inLib[1024];
    strcpy(inLib, sLibname.data());
    si2drErrorT err;
    si2drGroupsIdT groups;
    si2drStringT gtype;
    si2drGroupIdT group;
    time_t startt,endt;
    
    if( !inLib ) {
      printf("No filename specified!\n\n");
      exit(0);
    }
	
    si2drPIInit(&err);
    
    	
    cout << "\nBuild up the hierarchical tree through Liberty Parser......\n";
    time(&startt);
    	
    si2drReadLibertyFile(inLib,&err);
    cout << "\nLiberty Parser is finished......\n";
    si2drPISetNocheckMode(&err);
    si2drPISetDebugMode(&err);
	
    groups = si2drPIGetGroups(&err);
    		
    //while( !si2drObjectIsNull((group=si2drIterNextGroup(groups,&err)),&err) ) {
    if( !si2drObjectIsNull((group=si2drIterNextGroup(groups,&err)),&err) ) {
      si2drNamesIdT gnames;
      si2drStringT gname;
      
      gtype = si2drGroupGetGroupType(group,&err);
      string gType(gtype);
//    if ( gType != "library" ) continue;

      gnames = si2drGroupGetNames(group, &err);
      gname = si2drIterNextName(gnames,&err);
      name(gname); 
      si2drIterQuit(gnames,&err);
      	
      time(&startt);
//     cout << "\n\ntesting "<< inLib << "...\n";
      read(group, "");
      time(&endt);
     
      printf("Time used for reading liberty file is %d seconds\n", endt-startt);
			
    }
//    cout<<lineno<<" lines read in"<<endl;
    si2drIterQuit(groups,&err);
    si2drPIUnSetDebugMode(&err);
	
    si2drPIQuit(&err);

  }
  catch (ios_base::failure& var) {
    cerr << var.what()<<endl;
  }
}

void Library::WriteLibrary(const string& sLibname, bool bSorttiming)
{
  cout<<"Write library to '"<<sLibname<<"'"<<endl;
  Pin::bSort = bSorttiming;
  try {

    ofstream outLib(sLibname.data());
    cout << sLibname.data()<<endl;

    WriteTo(outLib, "");

    outLib.close();
  }
  catch (ios_base::failure& var) {
    cerr << var.what()<<endl;
  }
}

void Library::Split()
{
  map <string, size_t>::const_iterator pcell;

  for(pcell=cells.begin(); pcell!=cells.end(); pcell++) {
    Sublib(pcell->first, (pcell->first + ".lib").c_str());
  }
}

void Library::Sublib(const string& strCells, const char* libname)
{
  vector <string> lisCells;
  SplitStr(strCells, ' ', lisCells);

  try {

    vector <Entry *> Cells;
    for (int i=0; i<lisCells.size(); i++) {
      Cells.push_back(findSubgroup("cell", lisCells[i]));
    }

    cout<<"Write lib '"<<libname<<"'"<<endl;
    ofstream out(libname);

    out<<type()<<" ("<<name()<<") {"<<endl;

    // write attributes and other groups
    for(int i=0; i<entries.size();i++) {
      if (entries[i] != NULL && entries[i]->type() != "cell") {
        if (getClasstype(entries[i]->type()) == eTemplate) {
          continue;
        } else if (getClasstype(entries[i]->type()) == eScaling_factor) {
          bool bFind = false;
          for (int n=0; n<Cells.size(); n++) {
            if (entries[i]->name().find(Cells[n]->name()) == 0) { 
	      bFind = true; 
	      break;
	    }
          }
          if (bFind) 
            entries[i]->WriteTo(out, "\t");

           continue;
        } else {
          entries[i]->WriteTo(out, "\t");
        }
      }
    }
    // write templates
    map <string, int> lut;
    for(int i=0; i<Cells.size(); i++) {
      Cells[i]->searchTemplates(lut);
    }

    map <string, int>::const_iterator pt;
    for (pt=lut.begin(); pt!=lut.end(); pt++) {   
      Entry *entry = findTemplate(pt->first);
      entry->WriteTo(out, "\t");
    }

    // write cell group
    for(int i=0; i<Cells.size(); i++) {
      Cells[i]->WriteTo(out, "\t");
    }

    out<<"}"<<endl;
    out.close();
  }
  catch (const exception& var) {
    cerr << var.what()<<endl;
  }
}

void Library::Ccs2ecsm(const string& strVol, const string& strCap)
{
  Group::Ccs2ecsm(strVol, strCap);

  Entry * version=new Attr(0, "ecsm_version", "1.2", CLASS1);
  entries.insert(entries.begin(), version);
  Entry * define=new Attr(0, "define", "(ecsm_version, library, float)", CLASS2);
  entries.insert(entries.begin(), define);
  define=new Attr(0, "define", "(index_1, ecsm_capacitance, string)", CLASS2);
  entries.insert(entries.begin(), define);
  define=new Attr(0, "define_group", "(ecsm_capacitance, pin)", CLASS2);
  entries.insert(entries.begin(), define);
  define=new Attr(0, "define", "(values, ecsm_capacitance, string)", CLASS2);
  entries.insert(entries.begin(), define);
  define=new Attr(0, "define_group", "(ecsm_capacitance, fall_transition)", CLASS2);
  entries.insert(entries.begin(), define);
  define=new Attr(0, "define_group", "(ecsm_capacitance, rise_transition)", CLASS2);
  entries.insert(entries.begin(), define);
  define=new Attr(0, "define", "(values, ecsm_waveform, string)", CLASS2);
  entries.insert(entries.begin(), define);
  define=new Attr(0, "define", "(index_1, ecsm_waveform, string)", CLASS2);
  entries.insert(entries.begin(), define);
  define=new Attr(0, "define_group", "(ecsm_waveform, fall_transition)", CLASS2);
  entries.insert(entries.begin(), define);
  define=new Attr(0, "define_group", "(ecsm_waveform, rise_transition)", CLASS2);
  entries.insert(entries.begin(), define);

  removeSubgroup("output_current_");

  return;
}
