/******************************************************************************
    Copyright (c) 2007 Electronic Tools Company    ALL RIGHTS RESERVED

  The contents of this file are subject to the restrictions and limitations
  set forth in the SYNOPSYS Open Source License Version 1.0  (the "License"); 
  you may not use this file except in compliance with such restrictions 
  and limitations. You may obtain instructions on how to receive a copy of 
  the License at

  http://www.synopsys.com/partners/tapin/open_source/license_liberty.html. 

  Software distributed by Original Contributor under the License is 
  distributed on an "AS IS" basis, WITHOUT WARRANTY OF ANY KIND, either 
  expressed or implied. See the License for the specific language governing 
  rights and limitations under the License.

******************************************************************************/

#ifndef _REGEX_H__
#define _REGEX_H__

#include <regex.h>
#define MAXLEN 4096

enum {eGROUP_PAT = 0, 
      eLIB_PAT,
      eGROUP_END_PAT,
      eATTR_PAT1,
      eATTR_PAT2,
      eOPTION,
      eMAX_PAT};

struct Pattern_t
{
  const char *string;
  int nParam;
};

class Regex
{
  static bool bInit;
  static regex_t *preg[eMAX_PAT];
  static Pattern_t pattern[eMAX_PAT];

public:
  Regex();
  ~Regex();

  static bool Match(const char* str, int pindex, regmatch_t matchptr[])
  {
    if (bInit) init();
    int nmatch = (pattern[pindex].nParam == 0) ? 0 
                   : pattern[pindex].nParam+1;

    int rc = regexec(preg[pindex], str, nmatch, matchptr, REG_NOTEOL);
#ifdef _DEBUG
    switch(rc) {
      case REG_NOMATCH: break;
      case REG_ESPACE: cerr<<"REG_ESPACE"<<endl; break;
    }
#endif
    return rc==0;
  }

private:
  static void init() {
    for (int i=0; i<eMAX_PAT; i++) {
      preg[i] = new regex_t;
#ifdef _DEBUG
      int rc = regcomp(preg[i], pattern[i].string, REG_EXTENDED);
      switch(rc) {
        case REG_BADRPT: cerr<<"REG_BADRPT"<<endl; break;
        case REG_BADBR:  cerr<<"REG_BADBR"<<endl; break;
        case REG_EBRACE: cerr<<"REG_EBRACE"<<endl; break;
        case REG_EBRACK: cerr<<"REG_EBRACK"<<endl; break;
        case REG_ERANGE: cerr<<"REG_ERANGE"<<endl; break;
        case REG_ECTYPE: cerr<<"REG_ECTYPE"<<endl; break;
        case REG_ECOLLATE: cerr<<"REG_ECOLLATE"<<endl; break;
        case REG_EPAREN: cerr<<"REG_EPAREN"<<endl; break;
        case REG_ESUBREG: cerr<<"REG_ESUBREG"<<endl; break;
        case REG_EESCAPE: cerr<<"REG_EESCAPE"<<endl; break;
        case REG_BADPAT: cerr<<"REG_BADPAT"<<endl; break;
        case REG_ESPACE: cerr<<"REG_ESPACE"<<endl; break;
      }
#endif    
    }

    bInit = false;
  }

};

#endif

