/******************************************************************************
    Copyright (c) 2007 Electronic Tools Company    ALL RIGHTS RESERVED

  The contents of this file are subject to the restrictions and limitations
  set forth in the SYNOPSYS Open Source License Version 1.0  (the "License"); 
  you may not use this file except in compliance with such restrictions 
  and limitations. You may obtain instructions on how to receive a copy of 
  the License at

  http://www.synopsys.com/partners/tapin/open_source/license_liberty.html. 

  Software distributed by Original Contributor under the License is 
  distributed on an "AS IS" basis, WITHOUT WARRANTY OF ANY KIND, either 
  expressed or implied. See the License for the specific language governing 
  rights and limitations under the License.

******************************************************************************/

#include "stdc++.h"

#include "pin.h"
#include "value.h"
#include "group.h"
#include "attr.h"
#include "util.h"
#include <math.h>

bool Pin::bSort;

Pin::Pin(int lineno, const string& name, Group *p)
    : Group(lineno, "pin", name, p)
{
  bEcsm_cap=true;
}

Pin::~Pin()
{
}

void Pin::writeContents(ostream& os, const string& strPrefix)
{
  vector <string> conds;
  map <string, int> timings;

  for(int i=0; i<entries.size();i++) {
    if (entries[i] != NULL) {
      if (bSort && entries[i]->type() == "timing") {
        string key = entries[i]->key();
        timings[key] = i;
        conds.push_back(key);
      } else {
        entries[i]->WriteTo(os, strPrefix);
      }
    }
  }
    
  sort(conds.begin(), conds.end());
 
  for(int i=0; i<conds.size(); i++) {
    int idx = timings[conds[i]];
    entries[idx]->WriteTo(os, strPrefix);
  }
}


void Pin::Addecsm(const string& strCap, const string& direction)
{
  try
    {
      Group * receiver_cap=dynamic_cast<Group *>(findSubgroup("receiver_capacitance"));
      if (receiver_cap)
	{
	  Value * receiver_cap1=dynamic_cast<Value *>(receiver_cap->findSubgroup("receiver_capacitance1_"+direction));
	  Value * receiver_cap2=dynamic_cast<Value *>(receiver_cap->findSubgroup("receiver_capacitance2_"+direction));
	  if (receiver_cap1 && receiver_cap2)
	    {
	      vector < vector<double> > indexes;
	      string index_value;
	      Group *ecsm_cap=new Group(0, "ecsm_capacitance", direction, this);
	      entries.push_back(ecsm_cap);
	      
	      receiver_cap1->getIndexes(indexes);
	      for (int i=0; i<indexes[0].size(); ++i)
		{
		  char temp[20];
		  sprintf(temp, "%lf, ", indexes[0][i]);
		  index_value+=temp;
		}
	      index_value.replace(index_value.size()-2, 2, "");
	      Attr *index=new Attr(0, "index_1", index_value, CLASS1);
	      ecsm_cap->entries.push_back(index);
	      
	      vector <double> m_dblData1;
	      receiver_cap1->getData(m_dblData1);
	      vector <double> m_dblData2;
	      receiver_cap2->getData(m_dblData2);
	      string cap_value;
	      for (int i=0; i<indexes[0].size(); ++i)
		{
		  char temp[20];
		  double tmp;
		  if (strCap=="first")
		    tmp=m_dblData1[i];
		  else if (strCap=="second")
		    tmp=m_dblData2[i];
		  else if (strCap=="min")
		    tmp=my_fmin(m_dblData1[i], m_dblData2[i]);
		  else if (strCap=="max")
		    tmp=my_fmax(m_dblData1[i], m_dblData2[i]);
		  else if (strCap=="ave")
		    tmp=(m_dblData1[i]+m_dblData2[i])/2.0;

		  sprintf(temp, "%lf, ", tmp);
		  cap_value+=temp;
		}
	      cap_value.replace(cap_value.size()-2, 2, "");
	      
	      Attr *values=new Attr(0, "values", cap_value,CLASS1);
	      ecsm_cap->entries.push_back(values);
	    }
	}
    }
  catch (const exception& e)
    {
      int idxAttr=findAttr("direction");
      string inout=entries[idxAttr]->value();
      if ((parent()->type()!="test_cell") && (inout!="internal"))
	cout<<"Warning: cell: "<<parent()->name()<<"cell type: "<<parent()->type()<<" pin "<<name()<<" doesn't have a receiver_capacitance() group."<<endl;
    }

  return;
}
