/******************************************************************************
    Copyright (c) 2007 Electronic Tools Company    ALL RIGHTS RESERVED

  The contents of this file are subject to the restrictions and limitations
  set forth in the SYNOPSYS Open Source License Version 1.0  (the "License"); 
  you may not use this file except in compliance with such restrictions 
  and limitations. You may obtain instructions on how to receive a copy of 
  the License at

  http://www.synopsys.com/partners/tapin/open_source/license_liberty.html. 

  Software distributed by Original Contributor under the License is 
  distributed on an "AS IS" basis, WITHOUT WARRANTY OF ANY KIND, either 
  expressed or implied. See the License for the specific language governing 
  rights and limitations under the License.

******************************************************************************/

#include "stdc++.h"
#include "util.h"

#include "template.h"

Template::Template(int lineno, const string& type, const string& name, Group *p)
    : Group(lineno, type, name, p)
{
}

Template::~Template()
{
}

void Template::AfterRead()
{
  try {
    for(int i=1; i<=3; i++) {
      string str("variable_"); str += ('0'+i);

      size_t idxEntry = findAttr(str);
      vars.push_back(entries[idxEntry]->value());
      vars_index[entries[idxEntry]->value()] = i-1;
    }
  }
  catch(const exception&)
  {}

  try {
    for(int i=1; i<=3; i++) {
      string str("index_"); str += ('0'+i);

      size_t idxEntry = findAttr(str);

      string value = entries[idxEntry]->value();
      size_t pos;
      while((pos=value.find_first_of("()\"\\\n")) != string::npos) {
        value.erase(pos, 1);
      }

      vector<double> index;
      vector<int> index_precision;
      SplitStr(value, ',', index);
      SplitStr(value, ',', index_precision);
      indexes.push_back(index);
      indexes_precision.push_back(getPrecision(index_precision));

      delete entries[idxEntry];
      entries[idxEntry] = NULL;
    }
  }
  catch(const exception&)
  {}
}

int Template::findVarIndex(const string& name) const
{
  map <string, int>::const_iterator p = vars_index.find(name);
  if( p != vars_index.end()) {
    return p->second;
  } else {
    return -1;
  }
}

void Template::writeContents(ostream& os, const string& strPrefix)
{
  for (int iCol=0; iCol< vars.size(); iCol++) {
    os<<strPrefix<<"variable_"<<iCol+1<<" : "<<vars[iCol]<<';'<<endl;

    if (indexes.size() < iCol+1) continue;

    os<<strPrefix<<"index_"<<iCol+1<<"(\""<<JoinFixed(indexes[iCol], ',')
                                              <<"\");"<<endl;
  }
}
