/******************************************************************************
    Copyright (c) 2007 Electronic Tools Company    ALL RIGHTS RESERVED

  The contents of this file are subject to the restrictions and limitations
  set forth in the SYNOPSYS Open Source License Version 1.0  (the "License"); 
  you may not use this file except in compliance with such restrictions 
  and limitations. You may obtain instructions on how to receive a copy of 
  the License at

  http://www.synopsys.com/partners/tapin/open_source/license_liberty.html. 

  Software distributed by Original Contributor under the License is 
  distributed on an "AS IS" basis, WITHOUT WARRANTY OF ANY KIND, either 
  expressed or implied. See the License for the specific language governing 
  rights and limitations under the License.

******************************************************************************/

#include "stdc++.h"
#include <math.h>
#include "util.h"

double getUnit(string value)
{
  if ( value.find_first_of("p") != value.npos) {
    return 1e-12;
  } else if ( value.find_first_of("n") != value.npos) {
    return 1e-9;
  } else if ( value.find_first_of("u") != value.npos) {
    return 1e-6;
  } else if ( value.find_first_of("m") != value.npos) {
    return 1e-3;
  } else if ( value.find_first_of("c") != value.npos) {
    return 1e-2;
  } else if ( value.find_first_of("h") != value.npos) {
    return 1e+2;
  } else if ( value.find_first_of("k") != value.npos) {
    return 1e+3;
  } else if ( value.find_first_of("M") != value.npos) {
    return 1e+6;
  } else if ( value.find_first_of("G") != value.npos) {
    return 1e+9;
  } else if ( value.find_first_of("f") != value.npos) {
    return 1e-15;
  } else {
    return 1;
  }
}

double getTimeUnit(Group *lib)
{
    size_t idxEntry = lib->findAttr("time_unit");
    string value = lib->entries[idxEntry]->value();
    size_t pos;
    while((pos=value.find_first_of("()\"\\\n")) != string::npos) {
      value.erase(pos, 1);
    }
    return getUnit(value);
}

double getFactor(Group *lib)
{
  double time, voltage, current, capacitance;
  double factor;

  try {
    size_t idxEntry = lib->findAttr("capacitive_load_unit");
    string value = lib->entries[idxEntry]->value();
    size_t pos;
    while((pos=value.find_first_of("()\"\\\n")) != string::npos) {
      value.erase(pos, 1);
    }
    vector<string> index_string; 
    SplitStr(value, ',', index_string);

    if ( index_string.size() == 1 ) {
      capacitance = atof(index_string[0].data());
    } else {
      factor = atof(index_string[0].data());
      capacitance = getUnit(index_string[1]) * factor;
    }
      
    idxEntry = lib->findAttr("time_unit");
    value = lib->entries[idxEntry]->value();
    while((pos=value.find_first_of("()\"\\\n")) != string::npos) {
      value.erase(pos, 1);
    }
    time = getUnit(value);

    idxEntry = lib->findAttr("voltage_unit");
    value = lib->entries[idxEntry]->value();
    while((pos=value.find_first_of("()\"\\\n")) != string::npos) {
      value.erase(pos, 1);
    }
    voltage = getUnit(value);

    idxEntry = lib->findAttr("current_unit");
    value = lib->entries[idxEntry]->value();
    while((pos=value.find_first_of("()\"\\\n")) != string::npos) {
      value.erase(pos, 1);
    }
    current = getUnit(value);
  }
  
  catch   (const exception&) {}

  return capacitance * voltage /time/current;
  
}

int getPrecision(string value)
{
  int precision;
  size_t beg = value.find_first_of(".");
  size_t en = value.find_first_of("eE\n");
  if (en == value.npos) {
    en = value.size();
  } 
  if (beg == value.npos) {
    precision = 0;
  } else { 
    precision = en - beg -1;
  }
  return precision;
}

int getPrecision(vector <string>& value)
{
  int precision = getPrecision(value[0]);
  int current;

  for (int i=1; i<value.size(); i++) { 
    current = getPrecision(value[i]);
    if ( precision > current) {
      precision = current;
    }
  }
    
  return precision;
}

int getPrecision(int value)
{
  return value;
}

int getPrecision(vector <int>& value)
{
  int precision = value[0];
  int current;

  for (int i=1; i<value.size(); i++) { 
    current = value[i];
    if ( precision > current) {
      precision = current;
    }
  }
    
  return precision;
}

Group *findLibrary(Group *gp)
{
  Group *group = gp;

  while ( group->type() != "library") {
    group = group->parent();
  }
  return group;
}

Group *findCell(Group *gp)
{
  Group *group = gp;

  if (group->type() == "library") {
    return NULL;
  }
  while (group->type() != "library" &&  group->type() != "cell") {
    group = group->parent();
  }
  if (group->type() == "library") {
    return NULL;
  }
  return group;
}

Group *findPin(Group *gp)
{
  Group *group = gp;

  if (group->type() == "library") {
    return NULL;
  }
  while (group->type() != "library" && group->type() != "pin") {
    group = group->parent();
  }
  if (group->type() == "library") {
    return NULL;
  }
  return group;
}

bool isLibrary(Entry *gp)
{
  if (gp->type() == "library") {
    return true;
  } else {
    return false;
  }
}

bool isCell(Entry *gp)
{
  if (gp->type() == "cell") {
    return true;
  } else {
    return false;
  }
}

bool isVector(Entry *gp)
{
  if (gp->type() == "vector") {
    return true;
  } else {
    return false;
  }
}

int QuoteString(char *name)
{
  char *s = name;
  if( *s >= '0' && *s <= '9' )
    return 1;
  while( *s ) {
    if( *s == ' ' ) return 1;

    if( (*s < '0') || ( *s > '9' && *s < 'A' ) || (*s > 'Z' && *s < 'a') || (*s > 'z') ) {
	return 1;
    }
    s++;
  }
  return 0;
}

void SplitStr(const string& str, char ch, vector <double> &data)
{
  int ppos=0;
  int pos = str.find(ch);

  while(pos != string::npos) {
    data.push_back(atof(str.substr(ppos, pos-ppos).data()));
    ppos = pos+1;
    pos = str.find(ch, ppos);
  }

  data.push_back(atof(str.substr(ppos).data()));
}

void SplitStr(const string& str, char ch, vector <string> &data)
{
  int ppos=0;
  int pos = str.find(ch);

  while(pos != string::npos) {
    data.push_back(str.substr(ppos, pos-ppos));
    ppos = pos+1;
    pos = str.find(ch, ppos);
  }

  data.push_back(str.substr(ppos));
}

void SplitStr(const string& str, char ch, vector <int> &data)
{
  int ppos=0;
  int pos = str.find(ch);

  while(pos != string::npos) {
    data.push_back(getPrecision(str.substr(ppos, pos-ppos)));
    ppos = pos+1;
    pos = str.find(ch, ppos);
  }

  data.push_back(getPrecision(str.substr(ppos)));
}

string JoinFixed(const vector <double>& list, char ch)
{
  ostringstream osstr;

  osstr<<fixed<<list[0];
  for(int i=1; i<list.size(); i++) {
    osstr<<ch<<list[i];
  }

  return osstr.str();
}

string JoinScientific(const vector <double>& list, char ch)
{
  ostringstream osstr;

  osstr<<scientific<<list[0];
  for(int i=1; i<list.size(); i++) {
    osstr<<ch<<list[i];
  }

  return osstr.str();
}

int getIndex(const vector <double>& list, double v)
{
  if (v < list[0]) return -1;

  for (int i=0; i<list.size(); i++) {
    if (v==list[i]) {
      return i;
    }
  }

  return list.size()-1;
}

int getIndexLow(const vector <double>& list, double v, int *status)
{
  if (v > list[list.size()-1] + TOLERANCE) {
    *status = NONMATCHED;
    return list.size()-1;
  } else if ( v > list[list.size()-1] ){
    *status = MATCHED;
    return list.size()-1;
  }

  for (int i=list.size()-1; i>=0; i--) {
    if (v > list[i]) {
      if ( fabs(v-list[i+1]) < TOLERANCE) {
        *status = MATCHED;
      } else {
        *status = INTERPOLATION;
      }
      return i;
    }
  }

  if ( v < list[0] - TOLERANCE) { 
    *status = NONMATCHED;
  } else {
    *status = MATCHED;
  }

  return 0;
}

double linear_interpolation(double x, double x1, double y1, 
                            double x2, double y2)
{
  if (fabs(x1-x2) < 1e-6) {
    return (y1+y2)/2;
  } else  {
    return y1 + (x-x1)*(y2-y1)/(x2-x1);
  }
}

double my_fmin(double d1, double d2)
{
  if (d1>d2)
    return d2;
  else
    return d1;
}

double my_fmax(double d1, double d2)
{
  if (d1<d2)
    return d2;
  else
    return d1;
}

//~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
vector <double> calVoltage(vector <double>& times, 
                                      vector <double>& currents, 
				      double load, 
				      double v0,
				      double factor)
{
  vector <double> voltages;
  voltages.push_back(v0);

  for(int i=1; i<times.size(); i++) {
    double dt = times[i] - times[i-1];
    double dv = 0.5 / load * (currents[i] + currents[i-1]) * dt/factor;

    v0 += dv;
    voltages.push_back(v0);
  }

  return voltages;
}
