/******************************************************************************
    Copyright (c) 2007 Electronic Tools Company    ALL RIGHTS RESERVED

  The contents of this file are subject to the restrictions and limitations
  set forth in the SYNOPSYS Open Source License Version 1.0  (the "License"); 
  you may not use this file except in compliance with such restrictions 
  and limitations. You may obtain instructions on how to receive a copy of 
  the License at

  http://www.synopsys.com/partners/tapin/open_source/license_liberty.html. 

  Software distributed by Original Contributor under the License is 
  distributed on an "AS IS" basis, WITHOUT WARRANTY OF ANY KIND, either 
  expressed or implied. See the License for the specific language governing 
  rights and limitations under the License.

******************************************************************************/

#ifndef _UTIL_H__
#define _UTIL_H__
#include "group.h"
#include "library.h"

#define MATCHED 0
#define NONMATCHED 1
#define INTERPOLATION 2
#define TOLERANCE 1e-6 
#define TOLERANCE_TIME 1e-2 
#define STEPS 10 
#define LISTSIZE 5 

double getUnit(string value);
double getTimeUnit(Group *);
double getFactor(Group *);
int getPrecision(string value);
int getPrecision(vector <string>& value);
int getPrecision(int value);
int getPrecision(vector <int>& value);
Group *findLibrary(Group*);
Group *findCell(Group*);
Group *findPin(Group*);
bool isLibrary(Entry *);
bool isCell(Entry *);
bool isPin(Entry *);
bool isVector(Entry *);
int QuoteString(char *);
void SplitStr(const string& str, char ch, vector <double> &data);
void SplitStr(const string& str, char ch, vector <string> &data);
void SplitStr(const string& str, char ch, vector <int> &data);

string JoinFixed(const vector <double>&, char);
string JoinScientific(const vector <double>&, char);

int getIndex(const vector <double>&, double);
int getIndexLow(const vector <double>&, double, int *);
double linear_interpolation(double, double, double, double, double);

double my_fmin(double d1, double d2);
double my_fmax(double d1, double d2);

template <class T>
ostream& operator<<(ostream& os, const vector<T>& list)
{
  os<<list[0];
  for (int i=1; i<list.size(); i++) {
      os<<","<<list[i];
  }

  return os;
}
#endif

vector <double> calVoltage(vector <double>& times, 
				    vector <double>& currents, double load, double v0, double factor);

