/******************************************************************************
    Copyright (c) 2007 Electronic Tools Company    ALL RIGHTS RESERVED

  The contents of this file are subject to the restrictions and limitations
  set forth in the SYNOPSYS Open Source License Version 1.0  (the "License"); 
  you may not use this file except in compliance with such restrictions 
  and limitations. You may obtain instructions on how to receive a copy of 
  the License at

  http://www.synopsys.com/partners/tapin/open_source/license_liberty.html. 

  Software distributed by Original Contributor under the License is 
  distributed on an "AS IS" basis, WITHOUT WARRANTY OF ANY KIND, either 
  expressed or implied. See the License for the specific language governing 
  rights and limitations under the License.

******************************************************************************/

#include <math.h>

#include "stdc++.h"
#include "value.h"
#include "pin.h"
#include "util.h"

int Value::nGoodDelay=0;
int Value::nGoodSlew=0;
int Value::nVectors=0;
int Value::nIgnoreDelay=0;
int Value::nIgnoreSlew=0;

#define DRIVER 0
#define RECEIVER 1

Value::Value(int lineno, const string& type, const string& name, Group *p) 
   : Group(lineno, type, name, p), m_dblRefTime(0)
{
}

Value::~Value()
{
}

void Value::AfterRead()
{
  try {
    size_t idxEntry = findAttr("reference_time");
    string value = entries[idxEntry]->value();
    m_dblRefTime= atof(value.data());
    size_t pos;
    while((pos=value.find_first_of("()\"\\\n")) != string::npos) {
      value.erase(pos, 1);
    }
    m_dblRefTime_precision = getPrecision(value);
    delete entries[idxEntry];
    entries[idxEntry] = NULL;
  }
  catch(const exception&) {}

  try {
    for(int i=1; i<=3; i++) {
      string str("index_"); str += ('0'+i);

      size_t idxEntry = findAttr(str);
      string value = entries[idxEntry]->value();
      size_t pos;
      while((pos=value.find_first_of("()\"\\\n")) != string::npos) {
        value.erase(pos, 1);
      }

      vector<double> index; 
      vector<int> index_precision; 
      SplitStr(value, ',', index);
      SplitStr(value, ',', index_precision);
      indexes.push_back(index);
      indexes_precision.push_back(getPrecision(index_precision));

      delete entries[idxEntry];
      entries[idxEntry] = NULL;
    }
  }
  catch(const exception&) {}

  try {
    size_t idxEntry = findAttr("values");
    string value = entries[idxEntry]->value();
    size_t pos;
    while((pos=value.find_first_of("()\"\\\n")) != string::npos) {
            value.erase(pos,1);
    }

    vector<int> data_precision; 
    SplitStr(value, ',', m_dblData);
    SplitStr(value, ',', data_precision);
    m_dblData_precision = getPrecision(data_precision);

    delete entries[idxEntry];
    entries[idxEntry] = NULL;
  }
  catch(const exception&) {}
    
  if (indexes.empty()) {
//    if (name() == "scalar") {
//      cout << "This is scalar" << endl;
//    }
    try {
      Entry* entry = findTemplate(name());
      if (name() == "scalar") {
        cout << "scalar template found?" << endl;
      }
      entry->getIndexes(indexes);
      entry->getIndexes_precision(indexes_precision);
    }
    catch ( const exception& ) {}
  }
}

void Value::writeContents(ostream& os, const string& strPrefix)
{
  ostringstream osstr;

  osstr<<fixed;
  if (m_dblRefTime != 0) {
    osstr<<strPrefix<<"reference_time : "<<m_dblRefTime<<';'<<endl;
  } 

  for (int iCol=0; iCol< indexes.size(); iCol++) {
    osstr<<strPrefix<<"index_"<<iCol+1<<"(\""<<JoinFixed(indexes[iCol], ',')
                                   <<"\");"<<endl;
  }

  int nRows;
  if (indexes.size()>0) {
    nRows = indexes[indexes.size()-1].size();
  }
  int nCols = m_dblData.size() / nRows;

  osstr<<strPrefix<<"values(\""<<scientific;

  for (int iCol=0; iCol<nCols; iCol++) {
    if (iCol!=0) {
      osstr<<strPrefix<<"       \"";
    }

    vector <double> data(m_dblData.begin()+iCol*nRows, m_dblData.begin()+(iCol+1)*nRows);

    if (type() == "vector") {
      osstr<<JoinScientific(data, ',');
    } else {
      osstr<<JoinFixed(data, ',');
    }

    if (iCol != nCols-1) {
      osstr<<"\", \\"<<endl;
    }
  }

  osstr<<"\");"<<endl;

  for(int i=0; i<entries.size();i++) {
    if (entries[i] != NULL) {
      entries[i]->WriteTo(osstr, strPrefix);
    }
  }
  
  os<<osstr.str();
  
}

void Value::searchTemplates(map <string, int>& lut) const
{
  lut[name()] = 1;
}


void Value::Ccs2ecsm(const string& strVol, const string& strCap)
{
  if ((type() != "rise_transition") && (type() != "fall_transition"))
    return;
      
  Group *lib=findLibrary(this);
  Group *cell=findCell(this);
  Group *pin =findPin(this);

  if(indexes.size() == 2){
     for (int i=0; i<indexes[0].size(); ++i)
     {
        for (int j=0; j<indexes[1].size(); ++j)
        {
           add_ecsmwave(strVol, i, j);
        }
     }
  }
  
  int idxAttr=pin->findAttr("direction");
  string str=pin->entries[idxAttr]->value();
  Pin * outpin=dynamic_cast<Pin *>(pin);
  if ((str=="output") && outpin)
    outpin->bEcsm_cap=false;

  Group *p=parent();
  bool bPos=true;
  try
    {
      int idxAttr=p->findAttr("timing_sense");
      string str=p->entries[idxAttr]->value();
      if (str=="negative_unate")
	bPos=false;
    }
  catch (const exception& e)
    {

    }

  string gname;
  if (((type()=="rise_transition") && (bPos)) || ((type()=="fall_transition") && (!bPos)))
    gname="rise";
  else
    gname="fall";


  Value *r1, *r2;
  bool bTiming;
  try
    {
      r1=dynamic_cast<Value *>(p->findSubgroup("receiver_capacitance1_"+gname));
      r2=dynamic_cast<Value *>(p->findSubgroup("receiver_capacitance2_"+gname));
      bTiming=true;
    }
  catch (const exception &e)
    {
      r1=0;
      r2=0;
    }

  if (!r1 || !r2)
    {
      int idxAttr=p->findAttr("related_pin");
      string related_pin_name=p->entries[idxAttr]->value();
      if (Pin * inpin=dynamic_cast<Pin *>(cell->findSubgroup("pin", related_pin_name)))
	{
	  Entry * rev;
	  try
	    {
	      rev=inpin->findSubgroup("receiver_capacitance");
	    }
	  catch (const exception& e)
	    {
          return;
//	      throw runtime_error("Warning: No correponding ccs receiver_capacitance data in input .lib, for timing arc between "+related_pin_name+" -> "+pin->name());
	    }

	  if (Group* r=dynamic_cast<Group *>(rev))
	    {
	      try
		{
		  r1=dynamic_cast<Value *>(r->findSubgroup("receiver_capacitance1_"+gname));
		  r2=dynamic_cast<Value *>(r->findSubgroup("receiver_capacitance2_"+gname));
		}
	      catch (const exception &e)
		{
		  r1=0;
		  r2=0;
		}

	      if (!r1 || !r2)
		{
		  throw runtime_error("Warning: No correponding ccs receiver_capacitance1/2 data in input .lib, for timing arc between "+related_pin_name+" -> "+pin->name());
		}
	      else
		bTiming=false;
	    }
	  else
	    {
          return;
//	      throw runtime_error("Warning: No correponding ccs receiver_capacitance data in input .lib, for timing arc between "+related_pin_name+" -> "+pin->name());
	    }
	}
    }


  // fix C1
  if(indexes.size() == r1->indexes.size()){
     interp_cap(r1);
  }
   // fix C2
   if(indexes.size() == r2->indexes.size()){
      interp_cap(r2);
   }

  Entry * templat=lib->findTemplate(name());
  int load_index=templat->findVarIndex("total_output_net_capacitance");
  string cap_value;
  for (int i=0; i<indexes[0].size(); ++i)
    {
       if(indexes.size()==2){
      for (int j=0; j<indexes[1].size(); ++j)
	{
	  int k, l, size;
	  char stmp[20];
	  if (bTiming)
	    {
	      k=i;
	      l=j;
	      size=indexes[1].size();
	    }
	  else if (load_index)
	    {
	      k=i;
	      l=0;
	      size=1;
	    }
	  else
	    {
	      k=0;
	      l=j;
	      size=1;
	    }
     
     double tmp;
	  int idx=k*size+l;
	  if (strCap=="first")
	    tmp=r1->m_dblData[idx];
	  else if (strCap=="second")
	    tmp=r2->m_dblData[idx];
	  else if (strCap=="min")
	    tmp=my_fmin(r1->m_dblData[idx], r2->m_dblData[idx]);
	  else if (strCap=="max")
	    tmp=my_fmax(r1->m_dblData[idx], r2->m_dblData[idx]);
	  else if (strCap=="ave")
	    tmp=(r1->m_dblData[idx]+r2->m_dblData[idx])/2.0;

	  sprintf(stmp, "%lf, ", tmp);
	  cap_value+=stmp;
	}
       }
       else{
          double tmp;
     int idx;
     if(r1->indexes.size()==2)
	     idx=i*indexes[1].size();
     else
        idx=i;
	  if (strCap=="first")
	    tmp=r1->m_dblData[idx];
	  else if (strCap=="second")
	    tmp=r2->m_dblData[idx];
	  else if (strCap=="min")
	    tmp=my_fmin(r1->m_dblData[idx], r2->m_dblData[idx]);
	  else if (strCap=="max")
	    tmp=my_fmax(r1->m_dblData[idx], r2->m_dblData[idx]);
	  else if (strCap=="ave")
	    tmp=(r1->m_dblData[idx]+r2->m_dblData[idx])/2.0;

     char stmp[20];
	  sprintf(stmp, "%lf, ", tmp);
	  cap_value+=stmp;
       }
    }
  cap_value.replace(cap_value.size()-2, 2, "");

  Group *ecsm_cap=new Group(0, "ecsm_capacitance", gname, this);
  entries.push_back(ecsm_cap);

  Attr *values=new Attr(0, "values", cap_value,CLASS1);
  ecsm_cap->entries.push_back(values);

  try
    {
      int idxAttr=p->findAttr("related_pin");
      string related_pin_name=p->entries[idxAttr]->value();
      Entry * related_pin=cell->findSubgroup("pin", related_pin_name);
      if (Pin * inpin=dynamic_cast<Pin *>(related_pin))
	{
	  inpin->bEcsm_cap=false;
	}
    }
  catch (const exception& e)
    {

    }
  
  return;
}

void Value::add_ecsmwave(const string& strVol, const int& row, const int& col)
{
  bool bRise;
  if(type()=="rise_transition")
    bRise=true;
  else
    bRise=false;

  vector<double> user_vol;
  SplitStr(strVol, ',', user_vol);
  
  if (bRise)
    {
      sort(user_vol.begin(), user_vol.end());
    }
  else
    {
      sort(user_vol.begin(), user_vol.end());
      reverse(user_vol.begin(), user_vol.end());
    }
  
  Group * lib=findLibrary(this);
  Group * cell=findCell(this);
  Group * pin=findPin(this);
  size_t idxAttr=lib->findAttr("nom_voltage");
  double vdd=atof(lib->entries[idxAttr]->value().data());
  double vinit=bRise ? 0.0 : vdd;


  Group * ccs_current;
  string ccs_current_group;

  if (bRise)
    ccs_current_group="output_current_rise";
  else
    ccs_current_group="output_current_fall";

  try
    {
      ccs_current=dynamic_cast <Group *>(parent()->findSubgroup(ccs_current_group));
    }
  catch (const exception& e)
    {
      try
      {
	  Group *p=parent();
	  int idxAttr=p->findAttr("related_pin");
	  string inpin=p->entries[idxAttr]->value();
	  idxAttr=p->findAttr("timing_type");
	  string type=p->entries[idxAttr]->value();
	  if (type.find("three_state_")==string::npos)
	    cout<<"Warning: cell "<<cell->name()<<" No "<<ccs_current_group<<" under arc: "<<inpin<<"->"<<pin->name()<<" with timing_type: "<<type<<"."<<endl;
      }
      catch (const exception& e)
	{
	  
	}
      return;
    }

  unsigned vectorIdx=0;
  unsigned numVector=ccs_current->entries.size();
  unsigned nIdx0, nIdx1;
  if(indexes.size()==2){
     nIdx0=indexes[0].size();
     nIdx1=indexes[1].size();
     if(nIdx0*nIdx1 != numVector){
        return;
     }
     vectorIdx=row*nIdx1+col;
  }
  else if(indexes.size()==1){
     nIdx0=indexes[0].size();
     nIdx1=numVector/nIdx0;
     vectorIdx=row*nIdx1+nIdx0/2;
  }
  else
     return;

  Value* ccs_vector=dynamic_cast<Value *>(ccs_current->entries[vectorIdx]);

  Entry * templat=lib->findTemplate(name());
  int load_index=templat->findVarIndex("total_output_net_capacitance");

  double load=ccs_vector->indexes[load_index][0];

  vector<double> voltages=calVoltage(ccs_vector->indexes[2], ccs_vector->m_dblData, load, vinit, getFactor(lib));

  string value;
  vector<double> real_vol=user_vol;
  for (int i=0; i<real_vol.size(); ++i)
    {
      real_vol[i]*=vdd;
      int j;
      for (j=0; j<voltages.size(); ++j)
	{
	  if (bRise && (real_vol[i]>voltages[j]) && (real_vol[i]<voltages[j+1]))
	    break;
	  else if (!bRise && (real_vol[i]<voltages[j]) && (real_vol[i]>voltages[j+1]))
	    break;
	}
      double time=interp_time(load, ccs_vector->indexes[2][j], ccs_vector->indexes[2][j+1], ccs_vector->m_dblData[j], ccs_vector->m_dblData[j+1], real_vol[i]-voltages[j], getFactor(lib));
      char temp[20];
      sprintf(temp, "%lf, ", time);
      value+=temp;
    }

  value.replace(value.size()-2, 2, "");

  char gname[20];
  sprintf(gname, "\"%d\"", vectorIdx);
  Group *ecsm_wave=new Group(0, "ecsm_waveform", gname, this);
  char tmp[20];
  string strIdx;
  for (int ix=0; ix<user_vol.size(); ++ix)
    {
      sprintf(tmp, "%lf, ", user_vol[ix]);
      strIdx+=tmp;
    }
  strIdx.replace(strIdx.size()-2, 2, "");
  Attr *index1=new Attr(0, "index_1", strIdx, CLASS1);
  Attr *values=new Attr(0, "values", value, CLASS1);
  
  entries.push_back(ecsm_wave);
  ecsm_wave->entries.push_back(index1);
  ecsm_wave->entries.push_back(values);
  
  return;

}

double Value::interp_time(double load, double t1, double t2, double i1, double i2, double dvol, double factor)
{
  double time;

  if ((fabs(t2-t1)<1.0e-6) || (fabs(i2-i1)<1.0e-6))
    return (t1+t2)/2.0;
  else
    {
      time=(t2-t1)*(sqrt(i1*i1+2.0*load*dvol*factor*(i2-i1)/(t2-t1))-i1)/(i2-i1)+t1;
      if ((time>t1) && (time<t2))
	return time;
      else
	{
	  time=(t2-t1)*(-sqrt(i1*i1+2.0*load*dvol*factor*(i2-i1)/(t2-t1))-i1)/(i2-i1)+t1;
	  if ((time>t1) && (time<t2))
	    return time;
	  else
	    return (t1+t2)/2.0;
	}
    }
}

void Value::interp_cap(Value *r)
{
   std::vector<double> inCapT;
   if(1 == indexes.size()){
      unsigned nx=r->indexes[0].size();
      unsigned nxT=indexes[0].size();
      unsigned ix=0;
      for(size_t i=0; i<nxT; ++i){
         unsigned s=0;
         while(r->indexes[0][ix]-indexes[0][i]<1e-15 && ix<nx-1)
            ++ix;
         if(0 == ix)
            s=0;
         else
            s=ix-1;
         inCapT.push_back(r->m_dblData[s] +
                (indexes[0][i]-r->indexes[0][s])/(r->indexes[0][s+1]-r->indexes[0][s])
                *(r->m_dblData[s+1]-r->m_dblData[s]));
      }
   }
   else if(2 == indexes.size() && 2 == r->indexes.size()){
      Group *lib=findLibrary(this);
      Entry * templat=lib->findTemplate(name());
      int load_index=templat->findVarIndex("total_output_net_capacitance");
      Entry * capTmpl=lib->findTemplate(r->name());
      int capLoadIdx=capTmpl->findVarIndex("total_output_net_capacitance");
      unsigned nx, ny;
      unsigned nxT=indexes[0].size();
      unsigned nyT=indexes[1].size();
      vector<double> x, y;
      if(load_index == capLoadIdx){
         x=r->indexes[0];
         y=r->indexes[1];
         nx=x.size();
         ny=y.size();
      } else {
         x=r->indexes[1];
         y=r->indexes[0];
         nx=x.size();
         ny=y.size();
      }
      unsigned ix=0;
      for(size_t i=0; i<nxT; ++i){
         unsigned s=0;
         while(x[ix]-indexes[0][i]<1e-15 && ix<nx-1)
            ++ix;
         if(0 == ix)
            s=0;
         else
            s=ix-1;
         unsigned iy=0;
         for(size_t j=0; j<nyT; ++j){
            unsigned l=0;
            while(y[iy]-indexes[1][j]<1e-15 && iy<ny-1)
               ++iy;
            if(0 == iy)
               l=0;
            else
               l=iy-1;
            std::vector<double> line;
            std::vector<std::vector<double> > a;
            std::vector<double> b;
            for(unsigned p=0; p<4; ++p){
               line.clear();
               line.push_back(1.0);
               line.push_back(x[s+p/2]);
               line.push_back(y[l+p%2]);
               line.push_back(x[s+p/2] * y[l+p%2]);
               a.push_back(line);
               b.push_back(r->m_dblData[(s+p/2)*ny+(l+p%2)]);
            }
            std::vector<double> x=elimination(a, b);
            inCapT.push_back(1*x[0] + indexes[0][i]*x[1] + indexes[1][j]*x[2]
                            + indexes[0][i]*indexes[1][j]*x[3]);
         }
      }
   }
   r->m_dblData=inCapT;
   return;
}

std::vector<double>
Value::elimination(std::vector<std::vector<double> > a,
                   std::vector<double> b)
{
      std::vector<double> x;
 
   // check array
   unsigned n=a.size();
   if(n == 0) {
      cout << "no data in coefficient array" << endl;
      return x;
   }
   for(size_t i=0; i<n; ++i)
      if(a[i].size() != n){
         cout << "invalid size of coefficients in equation" << endl;
         return x;
      }
   if(b.size() != n){
      cout << "invalid size of coefficients in array b" << endl;
      return x;
   }
   
   // get triangle array
   std::vector<unsigned> idx;
   for(size_t i=0; i<n; ++i)
      idx.push_back(i);

   for(unsigned k=0; k<n; ++k){
      unsigned line=k;
      while(fabs(a[idx[line]][k]-0)<1e-15 && line<n){
         ++line;
      }
      if(line == n)
         continue;
      if(line != k){
         // exchange
         unsigned tmp=idx[k];
         idx[k]=idx[line];
         idx[line]=tmp;
      }

      // row
      for(unsigned i=k+1; i<n; ++i) {
         double pwr = a[idx[i]][k]/a[idx[k]][k];
         // col
         for(unsigned j=k; j<n; ++j)
            a[idx[i]][j] -= pwr*a[idx[k]][j];
         b[idx[i]] -= pwr*b[idx[k]];
      }
   }

   //
   for(int i=n-1; i>=0; --i){
      if(fabs(a[idx[i]][i]-0)<1e-15)
         b[idx[i]]=0.0;
      else {
         double pwr=0.0;
         for(int j=n-1; j>i; --j)
            pwr += a[idx[i]][j]*b[idx[j]];
         b[idx[i]] -= pwr;
         b[idx[i]] /= a[idx[i]][i];
      }
   }

   // collcet
   for(unsigned i=0; i<n; ++i)
      x.push_back(b[idx[i]]);
   return x;
}
